@@ -10,8 +10,10 @@
   Copyright 2016      Sjjad Hashemian
   Copyright 2016      Stephen Thirlwall
   Copyright 2017      Antonio Quartulli
-  Copyright 2017      Ismael Luceno
+  Copyright 2017-2019 Ismael Luceno
   Copyright 2017      nemermollon
+  Copyright 2018      Shankar
+  Copyright 2019      Evangelos Foutras
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
@@ -44,91 +46,83 @@
 
 /* Main control */
 
+#include "config.h"
 #include "axel.h"
 #include "assert.h"
+#include "sleep.h"
 
 /* Axel */
 static void save_state(axel_t *axel);
 static void *setup_thread(void *);
-static void axel_message(axel_t *axel, char *format, ...);
+
+#ifdef __GNUC__
+__attribute__((format(printf, 2, 3)))
+#endif /* __GNUC__ */
+static void axel_message(axel_t *axel, const char *format, ...);
 static void axel_divide(axel_t *axel);
 
 static char *buffer = NULL;
 
+#define MIN_CHUNK_WORTH (100 * 1024) /* 100 KB */
+
 /* Create a new axel_t structure */
 axel_t *
-axel_new(conf_t *conf, int count, const void *url)
+axel_new(conf_t *conf, int count, const search_t *res)
 {
-	const search_t *res;
 	axel_t *axel;
 	int status;
-	long delay;
+	uint64_t delay;
 	url_t *u;
 	char *s;
 	int i;
 
-	axel = malloc(sizeof(axel_t));
+	if (!count || !res)
+		return NULL;
+
+	axel = calloc(1, sizeof(axel_t));
 	if (!axel)
 		goto nomem;
 
-	memset(axel, 0, sizeof(axel_t));
 	axel->conf = conf;
-	axel->conn = malloc(sizeof(conn_t) * axel->conf->num_connections);
+	axel->conn = calloc(axel->conf->num_connections, sizeof(conn_t));
 	if (!axel->conn)
 		goto nomem;
-	memset(axel->conn, 0, sizeof(conn_t) * axel->conf->num_connections);
 
 	for (i = 0; i < axel->conf->num_connections; i++)
 		pthread_mutex_init(&axel->conn[i].lock, NULL);
 
 	if (axel->conf->max_speed > 0) {
-		if ((float)axel->conf->max_speed / axel->conf->buffer_size <
-		    0.5) {
+		/* max_speed / buffer_size < .5 */
+		if (16 * axel->conf->max_speed / axel->conf->buffer_size < 8) {
 			if (axel->conf->verbose >= 2)
 				axel_message(axel,
 					     _("Buffer resized for this speed."));
 			axel->conf->buffer_size = axel->conf->max_speed;
 		}
-		delay = (int)((float)1000000000 / axel->conf->max_speed *
-			      axel->conf->buffer_size *
-			      axel->conf->num_connections);
+		delay = 1000000000 * axel->conf->buffer_size *
+			axel->conf->num_connections / axel->conf->max_speed;
 
-		axel->delay_time.tv_sec = delay / 1000000000;
+		axel->delay_time.tv_sec  = delay / 1000000000;
 		axel->delay_time.tv_nsec = delay % 1000000000;
 	}
 	if (buffer == NULL) {
 		/* reserve 4 additional bytes for file extension ".st" */
-		buffer = malloc(max(MAX_STRING + 4, axel->conf->buffer_size));
+		buffer = malloc(max(MAX_STRING + 4,
+				(size_t)axel->conf->buffer_size));
 		if (!buffer)
 			goto nomem;
 	}
 
-	if (!url) {
-		axel_message(axel, _("Invalid URL"));
-		axel_close(axel);
-		return NULL;
-	}
+	u = malloc(sizeof(url_t) * count);
+	if (!u)
+		goto nomem;
+	axel->url = u;
 
-	if (count == 0) {
-		axel->url = malloc(sizeof(url_t));
-		if (!axel->url)
-			goto nomem;
-
-		axel->url->next = axel->url;
-		strncpy(axel->url->text, url, sizeof(axel->url->text) - 1);
-	} else {
-		res = url;
-		u = malloc(sizeof(url_t) * count);
-		if (!u)
-			goto nomem;
-		axel->url = u;
-
-		for (i = 0; i < count; i++) {
-			strncpy(u[i].text, res[i].url, sizeof(u[i].text) - 1);
-			u[i].next = &u[i + 1];
-		}
-		u[count - 1].next = u;
+	for (i = 0; i < count; i++) {
+		strlcpy(u[i].text, res[i].url, sizeof(u[i].text));
+		u[i].next = &u[i + 1];
 	}
+	u[count - 1].next = u;
 
 	axel->conn[0].conf = axel->conf;
 	if (!conn_set(&axel->conn[0], axel->url->text)) {
@@ -140,19 +134,21 @@ axel_new(conf_t *conf, int count, const void *url)
 	axel->conn[0].local_if = axel->conf->interfaces->text;
 	axel->conf->interfaces = axel->conf->interfaces->next;
 
-	strncpy(axel->filename, axel->conn[0].file, sizeof(axel->filename) - 1);
+	strlcpy(axel->filename, axel->conn[0].file, sizeof(axel->filename));
 	http_decode(axel->filename);
-	if (*axel->filename == 0)	/* Index page == no fn */
-		strncpy(axel->filename, axel->conf->default_filename,
-			sizeof(axel->filename) - 1);
+
 	if ((s = strchr(axel->filename, '?')) != NULL &&
 	    axel->conf->strip_cgi_parameters)
 		*s = 0;		/* Get rid of CGI parameters */
 
+	if (*axel->filename == 0)	/* Index page == no fn */
+		strlcpy(axel->filename, axel->conf->default_filename,
+			sizeof(axel->filename));
+
 	if (axel->conf->no_clobber && access(axel->filename, F_OK) == 0) {
 		char stfile[MAX_STRING + 3];
 
-		sprintf(stfile, "%s.st", axel->filename);
+		snprintf(stfile, sizeof(stfile), "%s.st", axel->filename);
 		if (access(stfile, F_OK) == 0) {
 			printf(_("Incomplete download found, ignoring "
 				 "no-clobber option\n"));
@@ -166,7 +162,7 @@ axel_new(conf_t *conf, int count, const void *url)
 
 	do {
 		if (!conn_init(&axel->conn[0])) {
-			axel_message(axel, axel->conn[0].message);
+			axel_message(axel, "%s", axel->conn[0].message);
 			axel->ready = -1;
 			return axel;
 		}
@@ -175,20 +171,19 @@ axel_new(conf_t *conf, int count, const void *url)
 		 * depends on the protocol used. */
 		status = conn_info(&axel->conn[0]);
 		if (!status) {
-			axel_message(axel, axel->conn[0].message);
+			axel_message(axel, "%s", axel->conn[0].message);
 			axel->ready = -1;
 			return axel;
 		}
-	}
-	/* re-init in case of protocol change. This can happen only once
-	 * because the FTP protocol can't redirect back to HTTP */
-	while (status == -1);
+	} while (status == -1); /* re-init in case of protocol change. This can
+				 * happen only once because the FTP protocol
+				 * can't redirect back to HTTP */
 
-	s = conn_url(axel->conn);
-	strncpy(axel->url->text, s, sizeof(axel->url->text) - 1);
+	conn_url(axel->url->text, sizeof(axel->url->text) - 1, axel->conn);
+	axel->size = axel->conn[0].size;
 	if (axel->conf->verbose > 0) {
-		if ((axel->size = axel->conn[0].size) != LLONG_MAX) {
-			axel_message(axel, _("File size: %lld bytes"),
+		if (axel->size != LLONG_MAX) {
+			axel_message(axel, _("File size: %zu bytes"),
 				     axel->size);
 		} else {
 			axel_message(axel, _("File size: unavailable"));
@@ -197,12 +192,12 @@ axel_new(conf_t *conf, int count, const void *url)
 
 	/* Wildcards in URL --> Get complete filename */
 	if (strchr(axel->filename, '*') || strchr(axel->filename, '?'))
-		strncpy(axel->filename, axel->conn[0].file,
-			sizeof(axel->filename) - 1);
+		strlcpy(axel->filename, axel->conn[0].file,
+			sizeof(axel->filename));
 
 	if (*axel->conn[0].output_filename != 0) {
-		strncpy(axel->filename, axel->conn[0].output_filename,
-			sizeof(axel->filename) - 1);
+		strlcpy(axel->filename, axel->conn[0].output_filename,
+			sizeof(axel->filename));
 	}
 
 	return axel;
@@ -258,10 +253,10 @@ axel_open(axel_t *axel)
 			return 0;
 		}
 
-		if (stsize < (sizeof(axel->conf->num_connections) +
-			      sizeof(axel->bytes_done) +
-			      2 * axel->conf->num_connections *
-			      sizeof(axel->conn[0].currentbyte))) {
+		if (stsize < (off_t)(sizeof(axel->conf->num_connections) +
+				     sizeof(axel->bytes_done) +
+				     2 * axel->conf->num_connections *
+				     sizeof(axel->conn[0].currentbyte))) {
 			/* FIXME this might be wrong, the file may have been
 			 * truncated, we need another way to check. */
 #ifdef DEBUG
@@ -298,7 +293,7 @@ axel_open(axel_t *axel)
 		}
 
 		axel_message(axel,
-			     _("State file found: %lld bytes downloaded, %lld to go."),
+			     _("State file found: %zu bytes downloaded, %zu to go."),
 			     axel->bytes_done, axel->size - axel->bytes_done);
 
 		close(fd);
@@ -337,8 +332,7 @@ axel_open(axel_t *axel)
 
 				if ((nwrite =
 				     write(axel->outfd, buffer,
-					   min(j,
-					       axel->conf->buffer_size))) < 0) {
+					   min(j, axel->conf->buffer_size))) < 0) {
 					if (errno == EINTR || errno == EAGAIN)
 						continue;
 					axel_message(axel,
@@ -353,6 +347,7 @@ axel_open(axel_t *axel)
 	return 1;
 }
 
+static
 void
 reactivate_connection(axel_t *axel, int thread)
 {
@@ -360,26 +355,26 @@ reactivate_connection(axel_t *axel, int thread)
 	int idx = -1;
 
 	if (axel->conn[thread].enabled ||
-	    axel->conn[thread].currentbyte <= axel->conn[thread].lastbyte)
+	    axel->conn[thread].currentbyte < axel->conn[thread].lastbyte)
 		return;
 	/* find some more work to do */
 	for (int j = 0; j < axel->conf->num_connections; j++) {
 		long long int remaining =
-		    axel->conn[j].lastbyte - axel->conn[j].currentbyte + 1;
+		    axel->conn[j].lastbyte - axel->conn[j].currentbyte;
 		if (remaining > max_remaining) {
 			max_remaining = remaining;
 			idx = j;
 		}
 	}
-	/* do not reactivate for less than 100KB */
-	if (max_remaining >= 100 * 1024 && idx != -1) {
+	/* do not reactivate unless large enough */
+	if (max_remaining > MIN_CHUNK_WORTH && idx != -1) {
 #ifdef DEBUG
 		printf(_("\nReactivate connection %d\n"), thread);
 #endif
 		axel->conn[thread].lastbyte = axel->conn[idx].lastbyte;
 		axel->conn[idx].lastbyte =
 		    axel->conn[idx].currentbyte + max_remaining / 2;
-		axel->conn[thread].currentbyte = axel->conn[idx].lastbyte + 1;
+		axel->conn[thread].currentbyte = axel->conn[idx].lastbyte;
 	}
 }
 
@@ -406,13 +401,10 @@ axel_start(axel_t *axel)
 	if (axel->conf->verbose > 0)
 		axel_message(axel, _("Starting download"));
 
-	for (i = 0; i < axel->conf->num_connections; i++)
+	for (i = 0; i < axel->conf->num_connections; i++) {
 		if (axel->conn[i].currentbyte > axel->conn[i].lastbyte) {
 			reactivate_connection(axel, i);
-		}
-
-	for (i = 0; i < axel->conf->num_connections; i++)
-		if (axel->conn[i].currentbyte <= axel->conn[i].lastbyte) {
+		} else if (axel->conn[i].currentbyte < axel->conn[i].lastbyte) {
 			if (axel->conf->verbose >= 2) {
 				axel_message(axel,
 					     _("Connection %i downloading from %s:%i using interface %s"),
@@ -429,9 +421,10 @@ axel_start(axel_t *axel)
 				axel->ready = -1;
 			}
 		}
+	}
 
 	/* The real downloading will start now, so let's start counting */
-	axel->start_time = gettime();
+	axel->start_time = axel_gettime();
 	axel->ready = 0;
 }
 
@@ -445,12 +438,12 @@ axel_do(axel_t *axel)
 	struct timeval timeval[1];
 	url_t *url_ptr;
 	struct timespec delay = {.tv_sec = 0, .tv_nsec = 100000000};
-	float max_speed_ratio;
+	unsigned int max_speed_ratio;
 
 	/* Create statefile if necessary */
-	if (gettime() > axel->next_state) {
+	if (axel_gettime() > axel->next_state) {
 		save_state(axel);
-		axel->next_state = gettime() + axel->conf->save_state_interval;
+		axel->next_state = axel_gettime() + axel->conf->save_state_interval;
 	}
 
 	/* Wait for data on (one of) the connections */
@@ -468,7 +461,7 @@ axel_do(axel_t *axel)
 	}
 	if (hifd == 0) {
 		/* No connections yet. Wait... */
-		if (axel_nanosleep(delay) < 0) {
+		if (axel_sleep(delay) < 0) {
 			axel_message(axel,
 				     _("Error while waiting for connection: %s"),
 				     strerror(errno));
@@ -499,7 +492,7 @@ axel_do(axel_t *axel)
 		if (!FD_ISSET(axel->conn[i].tcp->fd, fds)) {
 			time_t timeout = axel->conn[i].last_transfer +
 			    axel->conf->connection_timeout;
-			if (gettime() > timeout) {
+			if (axel_gettime() > timeout) {
 				if (axel->conf->verbose)
 					axel_message(axel,
 						     _("Connection %i timed out"),
@@ -509,7 +502,7 @@ axel_do(axel_t *axel)
 			goto next_conn;
 		}
 
-		axel->conn[i].last_transfer = gettime();
+		axel->conn[i].last_transfer = axel_gettime();
 		size =
 		    tcp_read(axel->conn[i].tcp, buffer,
 			     axel->conf->buffer_size);
@@ -525,7 +518,7 @@ axel_do(axel_t *axel)
 		if (size == 0) {
 			if (axel->conf->verbose) {
 				/* Only abnormal behaviour if: */
-				if (axel->conn[i].currentbyte <
+				if (axel->conn[i].currentbyte <=
 				    axel->conn[i].lastbyte &&
 				    axel->size != LLONG_MAX) {
 					axel_message(axel,
@@ -546,8 +539,7 @@ axel_do(axel_t *axel)
 		}
 
 		/* remaining == Bytes to go */
-		remaining =
-		    axel->conn[i].lastbyte - axel->conn[i].currentbyte + 1;
+		remaining = axel->conn[i].lastbyte - axel->conn[i].currentbyte;
 		if (remaining < size) {
 			if (axel->conf->verbose) {
 				axel_message(axel, _("Connection %i finished"),
@@ -607,14 +599,14 @@ axel_do(axel_t *axel)
 				if (pthread_create
 				    (axel->conn[i].setup_thread, NULL,
 				     setup_thread, &axel->conn[i]) == 0) {
-					axel->conn[i].last_transfer = gettime();
+					axel->conn[i].last_transfer = axel_gettime();
 				} else {
 					axel_message(axel,
 						     _("pthread error!!!"));
 					axel->ready = -1;
 				}
 			} else {
-				if (gettime() > (axel->conn[i].last_transfer +
+				if (axel_gettime() > (axel->conn[i].last_transfer +
 						 axel->conf->reconnect_delay)) {
 					pthread_cancel(*axel->conn[i].setup_thread);
 					axel->conn[i].state = false;
@@ -629,7 +621,7 @@ axel_do(axel_t *axel)
 	/* Calculate current average speed and finish_time */
 	axel->bytes_per_second =
 	    (int)((double)(axel->bytes_done - axel->start_byte) /
-		  (gettime() - axel->start_time));
+		  (axel_gettime() - axel->start_time));
 	if (axel->bytes_per_second != 0)
 		axel->finish_time =
 		    (int)(axel->start_time +
@@ -641,24 +633,22 @@ axel_do(axel_t *axel)
 	/* Check speed. If too high, delay for some time to slow things
 	   down a bit. I think a 5% deviation should be acceptable. */
 	if (axel->conf->max_speed > 0) {
-		max_speed_ratio = (float)axel->bytes_per_second /
+		max_speed_ratio = 1000 * axel->bytes_per_second /
 		    axel->conf->max_speed;
-		if (max_speed_ratio > 1.05)
+		if (max_speed_ratio > 1050) {
 			axel->delay_time.tv_nsec += 10000000;
-		else if ((max_speed_ratio < 0.95) &&
-			 (axel->delay_time.tv_nsec >= 10000000))
-			axel->delay_time.tv_nsec -= 10000000;
-		else if ((max_speed_ratio < 0.95) &&
-			 (axel->delay_time.tv_sec > 0)) {
-			axel->delay_time.tv_sec--;
-			axel->delay_time.tv_nsec += 999000000;
-		} else
-		    if (((float)axel->bytes_per_second / axel->conf->max_speed <
-			 0.95)) {
+		} else if (max_speed_ratio < 950) {
+			if (axel->delay_time.tv_nsec >= 10000000) {
+				axel->delay_time.tv_nsec -= 10000000;
+			} else if (axel->delay_time.tv_sec > 0) {
+				axel->delay_time.tv_sec--;
+				axel->delay_time.tv_nsec += 999000000;
+			}
+		} else {
 			axel->delay_time.tv_sec = 0;
 			axel->delay_time.tv_nsec = 0;
 		}
-		if (axel_nanosleep(axel->delay_time) < 0) {
+		if (axel_sleep(axel->delay_time) < 0) {
 			axel_message(axel,
 				     _("Error while enforcing throttling: %s"),
 				     strerror(errno));
@@ -714,15 +704,16 @@ axel_close(axel_t *axel)
 
 /* time() with more precision */
 double
-gettime()
+axel_gettime(void)
 {
 	struct timeval time[1];
 
-	gettimeofday(time, 0);
+	gettimeofday(time, NULL);
 	return (double)time->tv_sec + (double)time->tv_usec / 1000000;
 }
 
 /* Save the state of the current download */
+static
 void
 save_state(axel_t *axel)
 {
@@ -762,6 +753,7 @@ save_state(axel_t *axel)
 }
 
 /* Thread used to set up a connection */
+static
 void *
 setup_thread(void *c)
 {
@@ -774,9 +766,9 @@ setup_thread(void *c)
 
 	pthread_mutex_lock(&conn->lock);
 	if (conn_setup(conn)) {
-		conn->last_transfer = gettime();
+		conn->last_transfer = axel_gettime();
 		if (conn_exec(conn)) {
-			conn->last_transfer = gettime();
+			conn->last_transfer = axel_gettime();
 			conn->enabled = true;
 			goto out;
 		}
@@ -792,7 +784,7 @@ setup_thread(void *c)
 
 /* Add a message to the axel->message structure */
 static void
-axel_message(axel_t *axel, char *format, ...)
+axel_message(axel_t *axel, const char *format, ...)
 {
 	message_t *m;
 	va_list params;
@@ -800,11 +792,10 @@ axel_message(axel_t *axel, char *format, ...)
 	if (!axel)
 		goto nomem;
 
-	m = malloc(sizeof(message_t));
+	m = calloc(1, sizeof(message_t));
 	if (!m)
 		goto nomem;
 
-	memset(m, 0, sizeof(message_t));
 	va_start(params, format);
 	vsnprintf(m->text, MAX_STRING, format, params);
 	va_end(params);
@@ -830,25 +821,37 @@ axel_message(axel_t *axel, char *format, ...)
 static void
 axel_divide(axel_t *axel)
 {
-	int i;
+	/* Optimize the number of connections in case the file is small */
+	size_t maxconns = max(1u, axel->size / MIN_CHUNK_WORTH);
+	if (maxconns < axel->conf->num_connections)
+		axel->conf->num_connections = maxconns;
 
-	axel->conn[0].currentbyte = 0;
-	axel->conn[0].lastbyte = axel->size / axel->conf->num_connections - 1;
-	for (i = 1; i < axel->conf->num_connections; i++) {
-#ifdef DEBUG
-		printf(_("Downloading %lld-%lld using conn. %i\n"),
-		       axel->conn[i - 1].currentbyte,
-		       axel->conn[i - 1].lastbyte, i - 1);
-#endif
-		axel->conn[i].currentbyte = axel->conn[i - 1].lastbyte + 1;
-		axel->conn[i].lastbyte =
-		    axel->conn[i].currentbyte +
-		    axel->size / axel->conf->num_connections;
+	/* Calculate each segment's size */
+	size_t seg_len = axel->size / axel->conf->num_connections;
+
+	if (!seg_len) {
+		printf(_("Too few bytes remaining, forcing a single connection\n"));
+		axel->conf->num_connections = 1;
+		seg_len = axel->size;
+
+		conn_t *new_conn = realloc(axel->conn, sizeof(*axel->conn));
+		if (new_conn)
+			axel->conn = new_conn;
 	}
-	axel->conn[axel->conf->num_connections - 1].lastbyte = axel->size - 1;
+
+	for (int i = 0; i < axel->conf->num_connections; i++) {
+		axel->conn[i].currentbyte = seg_len * i;
+		axel->conn[i].lastbyte    = seg_len * i + seg_len;
+	}
+
+	/* Last connection downloads remaining bytes */
+	size_t tail = axel->size % seg_len;
+	axel->conn[axel->conf->num_connections - 1].lastbyte += tail;
 #ifdef DEBUG
-	printf(_("Downloading %lld-%lld using conn. %i\n"),
-	       axel->conn[i - 1].currentbyte, axel->conn[i - 1].lastbyte,
-	       i - 1);
+	for (int i = 0; i < axel->conf->num_connections; i++) {
+		printf(_("Downloading %lld-%lld using conn. %i\n"),
+		       axel->conn[i].currentbyte,
+		       axel->conn[i].lastbyte, i);
+	}
 #endif
 }
