@@ -7,7 +7,7 @@
   Copyright 2015-2017 Joao Eriberto Mota Filho
   Copyright 2016      Stephen Thirlwall
   Copyright 2017      Antonio Quartulli
-  Copyright 2017      Ismael Luceno
+  Copyright 2017-2019 Ismael Luceno
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
@@ -43,16 +43,12 @@
 #ifndef AXEL_AXEL_H
 #define AXEL_AXEL_H
 
-#include "config.h"
-
 #include <time.h>
 #include <ctype.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <stdio.h>
-#include <netdb.h>
 #ifndef	NOGETOPTLONG
-#define _GNU_SOURCE
 #include <getopt.h>
 #endif
 #include <limits.h>
@@ -64,27 +60,27 @@
 #include <stdbool.h>
 #include <sys/stat.h>
 #include <sys/time.h>
-#include <sys/types.h>
-#include <sys/ioctl.h>
-#include <sys/socket.h>
 #include <netinet/in_systm.h>
 #include <netinet/in.h>
 #include <netinet/ip.h>
 #include <arpa/inet.h>
 #include <net/if.h>
 #include <pthread.h>
-#ifdef HAVE_SSL
-#include <openssl/ssl.h>
-#endif
+#include "compat-android.h"
+#include "compat-bsd.h"
+#include "compat-ssl.h"
 
 /* Internationalization */
-#define PACKAGE			"axel"
+#ifdef ENABLE_NLS
 #define _(x)			gettext(x)
 #include <libintl.h>
 #include <locale.h>
+#else
+#define _(x)			(x)
+#endif
 
 /* Compiled-in settings */
-#define MAX_STRING		1024
+#define MAX_STRING		((size_t)1024)
 #define MAX_ADD_HEADERS	10
 #define MAX_REDIRECT		20
 #define DEFAULT_IO_TIMEOUT	120
@@ -106,8 +102,18 @@ typedef message_t if_t;
 #include "ssl.h"
 #include "search.h"
 
-#define min(a, b)		((a) < (b) ? (a) : (b))
-#define max(a, b)		((a) > (b) ? (a) : (b))
+#define min(a, b) \
+	({ \
+		__typeof__(a) __a = (a);	\
+		__typeof__(b) __b = (b);	\
+		__a < __b ? __a : __b;		\
+	})
+#define max(a, b) \
+	({ \
+		__typeof__(a) __a = (a);	\
+		__typeof__(b) __b = (b);	\
+		__a > __b ? __a : __b;		\
+	})
 
 typedef struct {
 	conn_t *conn;
@@ -115,7 +121,7 @@ typedef struct {
 	char filename[MAX_STRING];
 	double start_time;
 	int next_state, finish_time;
-	long long bytes_done, start_byte, size;
+	size_t bytes_done, start_byte, size;
 	int bytes_per_second;
 	struct timespec delay_time;
 	int outfd;
@@ -124,21 +130,16 @@ typedef struct {
 	url_t *url;
 } axel_t;
 
-axel_t *axel_new(conf_t *conf, int count, const void *url);
+axel_t *axel_new(conf_t *conf, int count, const search_t *urls);
 int axel_open(axel_t *axel);
 void axel_start(axel_t *axel);
 void axel_do(axel_t *axel);
 void axel_close(axel_t *axel);
 void print_messages(axel_t *axel);
 
-double gettime();
+double axel_gettime(void);
 
-static inline int
-axel_nanosleep(struct timespec delay)
-{
-	int res;
-	while ((res = nanosleep(&delay, &delay)) && errno == EINTR) ;
-	return res;
-}
+#define DN_MATCH_MALFORMED -1
+int dn_match(const char *hostname, const char *pat, size_t pat_len);
 
 #endif				/* AXEL_AXEL_H */
